package test;

/*
 *  Copyright (C) 2020 Juergen Platzer
 *
 *  This file is part of JInfoViz.
 *
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */


import tools.SquareMatrix;

public abstract class AbstractTest {

  public static final String TEST_SEPARATOR = "==================================================";
  public static final String      SEPARATOR = "--------------------------------------------------";
  public static final String FAILED = "[FAILED] ";
  public static final String OK = "[OK] ";

  public static final double DEFAULT_DOUBLE_ALLOWED_DIFFERENCE = 0.0000001;

  private int _errorCount;



  public void performTest() {
    System.out.println(TEST_SEPARATOR);
    System.out.println(getTestName());
    System.out.println(SEPARATOR);

    performActualTest();

    printTestSummary();

    System.out.println(TEST_SEPARATOR);
    System.out.println("");


  }

  public int getErrorCount() {
    return _errorCount;
  }


  ///////////////////////////////////////////////
  // UTILITY METHODS FOR THE TEST CLASSES      //
  ///////////////////////////////////////////////

  protected void initTest() {
    _errorCount = 0;
  }

  protected void performTestComparison(String testInformation, boolean expectedValue, boolean actualValue) {
    if (expectedValue == actualValue) {
      printOkText(testInformation);
    }
    else {
      handleError(testInformation, getBooleanDifferenceText(expectedValue, actualValue));
    }
  }

  protected void performTestComparison(String testInformation, double expectedValue, double actualValue) {
    performTestComparison(testInformation, expectedValue, actualValue, 0);
  }

  protected void performTestComparison(String testInformation, double expectedValue, double actualValue, double allowedDifference) {
    if (isDoubleValueOk(expectedValue, actualValue, allowedDifference)) {
      printOkText(testInformation);
    }
    else {
      handleError(testInformation, getDoubleDifferenceText(expectedValue, actualValue));
    }
  }


  protected void performTestComparison(String testInformation, SquareMatrix expectedValue, SquareMatrix actualValue, double allowedDifference) {
    if (expectedValue == null) {
      handleError(testInformation, "Expected matrix is null");
      return;
    }

    if (actualValue == null) {
      handleError(testInformation, "Actual matrix is null");
      return;
    }


    double[][] expected = expectedValue.getCopyOfValues();
    double[][] actual = actualValue.getCopyOfValues();
    int expectedDimension = expected.length;

    if (expectedDimension != actual.length) {
      String error = "Dimensionality does not match: expected value = " + expectedDimension + " vs. actual value = " + actual.length;
      handleError(testInformation, error);
      return;
    }


    boolean ok = true;
    for (int i = 0; i < expectedDimension; i++) {
      for (int j = 0; j < expectedDimension; j++) {
        if (!isDoubleValueOk(expected[i][j], actual[i][j], allowedDifference)) {

          String error = "Matrix value [" + i + "][" + j + "] not matching: " + getDoubleDifferenceText(expected[i][j], actual[i][j]);
          handleError(testInformation, error);
          return;
        }
      }
    }




    printOkText(testInformation);

  }


  protected void performTestComparison(String testInformation, double[] expectedVector, double[] actualVector, double allowedDifference) {
    if (expectedVector == null) {
      handleError(testInformation, "Expected vector is null");
      return;
    }

    if (actualVector == null) {
      handleError(testInformation, "Actual vector is null");
      return;
    }

    int expectedDimension = expectedVector.length;

    if (expectedDimension != actualVector.length) {
      String error = "Dimensionality does not match: expected value = " + expectedDimension + " vs. actual value = " + actualVector.length;
      handleError(testInformation, error);
      return;
    }

    //find an error with the current vector settings
    String error = null;
    //actual vector
    for (int i = 0; i < expectedDimension; i++) {
      //actual and inverted vector
      if (!isDoubleValueOk(expectedVector[i], actualVector[i], allowedDifference)) {
        error = "Expected value [" + i + "] not matching: " + getDoubleDifferenceText(expectedVector[i], actualVector[i]);
        break;
      }
    }

    //if there are also errors with the inverted vector -> print the error that was found previously.
    if (error != null) {
      for (int i = 0; i < expectedDimension; i++) {
        if (!isDoubleValueOk(expectedVector[i], -1*actualVector[i], allowedDifference)) {
          handleError(testInformation, error);
          return;

        }
      }
    }

    printOkText(testInformation);


  }

  protected void handleError(String testInformation, String error) {
    _errorCount++;
    System.out.println(FAILED + testInformation + ": " + error);
  }


  protected abstract String getTestName();

  protected abstract void performActualTest();



  ////////////////////////////////
  // PRIVATE METHODS            //
  ////////////////////////////////
  private boolean isDoubleValueOk(double expectedValue, double actualValue, double allowedDifference) {
    return Math.abs(expectedValue - actualValue) <= Math.abs(allowedDifference);
  }

  private String getDoubleDifferenceText(double expectedValue, double actualValue) {
    String differenceText = "expected value = " + expectedValue + " vs. actual value = " + actualValue;
    differenceText += " (Difference: " + (expectedValue - actualValue) + ")";
    return differenceText;
  }

  private String getBooleanDifferenceText(boolean expectedValue, boolean actualValue) {
    String differenceText = "expected value = " + expectedValue + " vs. actual value = " + actualValue;
    return differenceText;
  }


  private void printTestSummary() {
    System.out.println(SEPARATOR);
    if (_errorCount == 0) {
      printOkText(getTestName());
    }
    else {
      System.out.println(FAILED + _errorCount + " ERRORS");
    }
  }

  private void printOkText(String testInformation) {
     System.out.println(OK + testInformation);
  }






}
