package test;

/*
 *  Copyright (C) 2020 Juergen Platzer
 *
 *  This file is part of JInfoViz.
 *
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */


import tools.EigenvalueDecomposition;
import tools.EigenValueAndVector;
import tools.SymmetricMatrix;
import java.util.Vector;

public class EigenvalueDecompositionTest extends AbstractTest {



  public EigenvalueDecompositionTest() {
    initTest();
  }


  protected String getTestName() {
    return "Eigenvalue DecompositionTest";
  }



  protected void performActualTest() {
    testCase1();
    testCase2();
    testIris();
  }


  ////////////////////////////////////////
  // PRIVATE METHODS                    //
  ////////////////////////////////////////
  /*
  R code:

  A <- matrix( c ( 1, 2, 2, 1 ), nrow=2, byrow=TRUE)
  1 2
  2 1

  eigen(A)
  eigen() decomposition
  $values
  [1]  3 -1

  $vectors
            [,1]       [,2]
  [1,] 0.7071068 -0.7071068
  [2,] 0.7071068  0.7071068
  */
  private void testCase1() {
    SymmetricMatrix matrix = new SymmetricMatrix(2);
    matrix.setValue(0,1,2);

    Vector<EigenValueAndVector> evdResult = null;
    try {
      EigenvalueDecomposition evd = new EigenvalueDecomposition(matrix);
      evdResult = evd.getEigenValuesAndVectors();
    }
    catch (Exception e) {
      e.printStackTrace();
    }

    if (evdResult.size() != 2) {
      handleError("Eigenvalue Decomposition Test 1", "Not the correct number of Eigenvalues detected!");
      return;
    }

    performTestComparison("Eigenvalue Decomposition Test 1 - Eigenvalue 1", 3, evdResult.get(0).getValue(), DEFAULT_DOUBLE_ALLOWED_DIFFERENCE);
    performTestComparison("Eigenvalue Decomposition Test 1 - Eigenvalue 2", 1, evdResult.get(1).getValue(), DEFAULT_DOUBLE_ALLOWED_DIFFERENCE);

    performTestComparison("Eigenvalue Decomposition Test 1 - Eigenvector 1", new double[]{0.7071068, 0.7071068}, evdResult.get(0).getVector(), DEFAULT_DOUBLE_ALLOWED_DIFFERENCE);
    performTestComparison("Eigenvalue Decomposition Test 1 - Eigenvector 2", new double[]{-0.7071068, 0.7071068}, evdResult.get(1).getVector(), DEFAULT_DOUBLE_ALLOWED_DIFFERENCE);


  }


  /*
  B <- matrix( c ( 15.4355238, 0.1076699, 0.1076699, 1.0324702 ), nrow=2, byrow=TRUE)

  eigen(B)
  eigen() decomposition
  $values
  [1] 15.436329  1.031665

  $vectors
               [,1]         [,2]
  [1,] -0.999972063  0.007474865
  [2,] -0.007474865 -0.999972063
  */


  private void testCase2() {
    SymmetricMatrix matrix = new SymmetricMatrix(2);
    matrix.setValue(0,0,15.4355238);
    matrix.setValue(0,1,0.1076699);
    matrix.setValue(1,1,1.0324702);

   Vector<EigenValueAndVector> evdResult = null;
    try {
      EigenvalueDecomposition evd = new EigenvalueDecomposition(matrix);
      evdResult = evd.getEigenValuesAndVectors();
    }
    catch (Exception e) {
      e.printStackTrace();
    }

    if (evdResult.size() != 2) {
      handleError("Eigenvalue Decomposition Test 2", "Not the correct number of Eigenvalues detected!");
      return;
    }

    performTestComparison("Eigenvalue Decomposition Test 2 - Eigenvalue 1", 15.436329, evdResult.get(0).getValue(), 0.000001);
    performTestComparison("Eigenvalue Decomposition Test 2 - Eigenvalue 2", 1.031665, evdResult.get(1).getValue(), 0.000001);

    performTestComparison("Eigenvalue Decomposition Test 2 - Eigenvector 1", new double[]{-0.999972063, -0.007474865}, evdResult.get(0).getVector(), DEFAULT_DOUBLE_ALLOWED_DIFFERENCE);
    performTestComparison("Eigenvalue Decomposition Test 2 - Eigenvector 2", new double[]{0.007474865, -0.999972063}, evdResult.get(1).getVector(), DEFAULT_DOUBLE_ALLOWED_DIFFERENCE);

  }



  /**


  cov(iris[1:4])
                 Sepal.Length Sepal.Width Petal.Length Petal.Width
    Sepal.Length    0.6856935  -0.0424340    1.2743154   0.5162707
    Sepal.Width    -0.0424340   0.1899794   -0.3296564  -0.1216394
    Petal.Length    1.2743154  -0.3296564    3.1162779   1.2956094
    Petal.Width     0.5162707  -0.1216394    1.2956094   0.5810063
    > C = cov(iris[1:4])
    eigen(C)



    eigen() decomposition
    $values
    [1] 4.22824171 0.24267075 0.07820950 0.02383509

    $vectors
                [,1]        [,2]        [,3]       [,4]
    [1,]  0.36138659 -0.65658877 -0.58202985  0.3154872
    [2,] -0.08452251 -0.73016143  0.59791083 -0.3197231
    [3,]  0.85667061  0.17337266  0.07623608 -0.4798390
    [4,]  0.35828920  0.07548102  0.54583143  0.7536574

  */



  private void testIris() {
    SymmetricMatrix matrix = new SymmetricMatrix(4);
    matrix.setValue(0,0,0.6856935);
    matrix.setValue(0,1,-0.0424340);
    matrix.setValue(0,2,1.2743154);
    matrix.setValue(0,3,0.5162707);
    matrix.setValue(1,1,0.1899794);
    matrix.setValue(1,2,-0.3296564);
    matrix.setValue(1,3,-0.1216394);
    matrix.setValue(2,2,3.1162779);
    matrix.setValue(2,3,1.2956094);
    matrix.setValue(3,3,0.5810063);


    Vector<EigenValueAndVector> evdResult = null;
    try {
      EigenvalueDecomposition evd = new EigenvalueDecomposition(matrix);
      evdResult = evd.getEigenValuesAndVectors();
    }
    catch (Exception e) {
      e.printStackTrace();
    }

    if (evdResult.size() != 4) {
      handleError("Iris covariance matrix Eigenvalue Decomposition", "Not the correct number of Eigenvalues detected!");
      return;
    }

    performTestComparison("Iris covariance matrix Eigenvalue Decomposition - Eigenvalue 1", 4.22824171, evdResult.get(0).getValue(), 0.000001);
    performTestComparison("Iris covariance matrix Eigenvalue Decomposition - Eigenvalue 2", 0.24267075, evdResult.get(1).getValue(), 0.000001);
    performTestComparison("Iris covariance matrix Eigenvalue Decomposition - Eigenvalue 3", 0.07820950, evdResult.get(2).getValue(), 0.000001);
    performTestComparison("Iris covariance matrix Eigenvalue Decomposition - Eigenvalue 4", 0.02383509, evdResult.get(3).getValue(), 0.000001);

    performTestComparison("Iris covariance matrix Eigenvalue Decomposition - Eigenvector 1", new double[]{0.36138659, -0.08452251, 0.85667061, 0.35828920}, evdResult.get(0).getVector(), DEFAULT_DOUBLE_ALLOWED_DIFFERENCE);
    performTestComparison("Iris covariance matrix Eigenvalue Decomposition - Eigenvector 2", new double[]{-0.65658877, -0.73016143, 0.17337266, 0.07548102}, evdResult.get(1).getVector(), DEFAULT_DOUBLE_ALLOWED_DIFFERENCE);
    performTestComparison("Iris covariance matrix Eigenvalue Decomposition - Eigenvector 3", new double[]{-0.58202985, 0.59791083, 0.07623608, 0.54583143}, evdResult.get(2).getVector(), 0.000001);
    performTestComparison("Iris covariance matrix Eigenvalue Decomposition - Eigenvector 4", new double[]{0.3154872, -0.3197231, -0.4798390, 0.7536574}, evdResult.get(3).getVector(), 0.000001);
  }

}
