package test;

/*
 *  Copyright (C) 2020 Juergen Platzer
 *
 *  This file is part of JInfoViz.
 *
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */

import tools.SymmetricMatrix;
import tools.SquareMatrix;
import java.util.Random;

public class MatrixInversionTest extends AbstractTest {



  public MatrixInversionTest() {
    initTest();
  }


  protected String getTestName() {
    return "Matrix Inversion Test";
  }



  protected void performActualTest() {
    testCase1();
    testCase2();
    testIris();
    testNotInvertable1();
  }


  ////////////////////////////////////////
  // PRIVATE METHODS                    //
  ////////////////////////////////////////
  /*
  R code:

  A <- matrix( c ( 1, 2, 2, 1 ), nrow=2, byrow=TRUE)
  1 2
  2 1

  solve(A)
             [,1]       [,2]
  [1,] -0.3333333  0.6666667
  [2,]  0.6666667 -0.3333333
        0.0000001

  det(A)

  = -3


   A %*% solve(A)
       [,1] [,2]
  [1,]    1    0
  [2,]    0    1
  */
  private void testCase1() {
    SymmetricMatrix matrix = new SymmetricMatrix(2);
    matrix.setValue(0,1,2);

    SquareMatrix actualInverse = null;
    double determinant = 0;
    try {
      actualInverse = matrix.getInverse();
      determinant = matrix.getDeterminant();
      performTestComparison("Matrix inversion 1 (invertable)", false, matrix.isNotInvertable());
    }
    catch (Exception e) {
      e.printStackTrace();
    }

    SymmetricMatrix expectedInverse = new SymmetricMatrix(2);
    expectedInverse.setValue(0, 0, -0.3333333);
    expectedInverse.setValue(0, 1, 0.6666667);
    expectedInverse.setValue(1, 1, -0.3333333);


    performTestComparison("Matrix inversion 1", expectedInverse, actualInverse, DEFAULT_DOUBLE_ALLOWED_DIFFERENCE);
    performTestComparison("Matrix determinant 1", -3, determinant, DEFAULT_DOUBLE_ALLOWED_DIFFERENCE);
  }


  /*
  B <- matrix( c ( 15.4355238, 0.1076699, 0.1076699, 1.0324702 ), nrow=2, byrow=TRUE)

  solve(B)
                [,1]         [,2]
  [1,]  0.064832782 -0.006761008
  [2,] -0.006761008  0.969256020


    det(B)
  [1] 15.92513

  > B%*%solve(B)
             [,1] [,2]
[1,] 1.000000e+00    0
[2,] 8.673617e-19    1

  */


  private void testCase2() {
    SymmetricMatrix matrix = new SymmetricMatrix(2);
    matrix.setValue(0,0,15.4355238);
    matrix.setValue(0,1,0.1076699);
    matrix.setValue(1,1,1.0324702);

    SquareMatrix actualInverse = null;
    double determinant = 0;
    try {
      actualInverse = matrix.getInverse();
      determinant = matrix.getDeterminant();
      performTestComparison("Matrix inversion 2 (invertable)", false, matrix.isNotInvertable());
    }
    catch (Exception e) {
      e.printStackTrace();
    }

    SymmetricMatrix expectedInverse = new SymmetricMatrix(2);
    expectedInverse.setValue(0, 0, 0.064832782);
    expectedInverse.setValue(0, 1, -0.006761008);
    expectedInverse.setValue(1, 1, 0.969256020);


    performTestComparison("Matrix inversion 2", expectedInverse, actualInverse, DEFAULT_DOUBLE_ALLOWED_DIFFERENCE);
    performTestComparison("Matrix determinant 2", 15.92513, determinant, 0.00001);
  }



  /**
    cov(iris[1:4])
               Sepal.Length Sepal.Width Petal.Length Petal.Width
  Sepal.Length    0.6856935  -0.0424340    1.2743154   0.5162707
  Sepal.Width    -0.0424340   0.1899794   -0.3296564  -0.1216394
  Petal.Length    1.2743154  -0.3296564    3.1162779   1.2956094
  Petal.Width     0.5162707  -0.1216394    1.2956094   0.5810063
  > C = cov(iris[1:4])
  > det(C)
  [1] 0.00191273
  > solve(C)
               Sepal.Length Sepal.Width Petal.Length Petal.Width
  Sepal.Length    10.314699   -6.713189    -7.314483    5.739951
  Sepal.Width     -6.713189   11.058417     6.480589   -6.170932
  Petal.Length    -7.314483    6.480589    10.031679  -14.513767
  Petal.Width      5.739951   -6.170932   -14.513767   27.693635

  */



  private void testIris() {
    SymmetricMatrix matrix = new SymmetricMatrix(4);
    matrix.setValue(0,0,0.6856935);
    matrix.setValue(0,1,-0.0424340);
    matrix.setValue(0,2,1.2743154);
    matrix.setValue(0,3,0.5162707);
    matrix.setValue(1,1,0.1899794);
    matrix.setValue(1,2,-0.3296564);
    matrix.setValue(1,3,-0.1216394);
    matrix.setValue(2,2,3.1162779);
    matrix.setValue(2,3,1.2956094);
    matrix.setValue(3,3,0.5810063);

    SquareMatrix actualInverse = null;
    double determinant = 0;
    try {
      actualInverse = matrix.getInverse();
      determinant = matrix.getDeterminant();
      performTestComparison("Iris covariance matrix inversion (invertable)", false, matrix.isNotInvertable());
    }
    catch (Exception e) {
      e.printStackTrace();
    }

    SymmetricMatrix expectedInverse = new SymmetricMatrix(4);
    expectedInverse.setValue(0,0,10.314699);
    expectedInverse.setValue(0,1,-6.713189);
    expectedInverse.setValue(0,2,-7.314483);
    expectedInverse.setValue(0,3,5.739951);
    expectedInverse.setValue(1,1,11.058417);
    expectedInverse.setValue(1,2,6.480589);
    expectedInverse.setValue(1,3,-6.170932);
    expectedInverse.setValue(2,2,10.031679);
    expectedInverse.setValue(2,3,-14.513767);
    expectedInverse.setValue(3,3,27.693635);


    performTestComparison("Iris covariance matrix inversion", expectedInverse, actualInverse, 0.0001);
    performTestComparison("Iris covariance matrix determinant", 0.00191273, determinant, DEFAULT_DOUBLE_ALLOWED_DIFFERENCE);
  }


  /**
      x = c(1:10)
    > y = 2*x
    > A <- matrix(c(x,y), nrow=10, byrow=FALSE)
    > cov(A)
             [,1]     [,2]
    [1,]  9.166667 18.33333
    [2,] 18.333333 36.66667
    > det(cov(A))
    [1] 0

    solve(cov(A))
    Fehler in solve.default(cov(A)) :
      Lapackroutine dgesv: System ist genau singulär: U[2,2] = 0

      [java] Matrix (2x2) =
      [java] 222222,256   -111111,098
      [java] -111111,098   55555,561

      B <- matrix( c ( 222222.256, -111111.098, -111111.098, 55555.561 ), nrow=2, byrow=TRUE)

      cov(A)%*%B
           [,1] [,2]
      [1,] 0.55 0.22
      [2,] 1.10 0.44


**/
  private void testNotInvertable1() {
    SymmetricMatrix matrix = new SymmetricMatrix(2);
    matrix.setValue(0,0,9.166667);
    matrix.setValue(0,1,18.33333);
    matrix.setValue(1,1,36.66667);

    SquareMatrix actualInverse = null;
    double determinant = 0;
    try {
      //System.out.println(matrix.getDeterminant());
      performTestComparison("Not invertable test 1", true, matrix.isNotInvertable());
      actualInverse = matrix.getInverse();
      //actualInverse.printMatrixValues();
    }
    catch (Exception e) {
      e.printStackTrace();
    }

  }

}
