package tools;

/*
 *  Copyright (C) 2018-2020 Juergen Platzer
 *
 *  This file is part of JInfoViz.
 *
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */


//import java.util.*;



/** Class implementing the classic and robust calculation of the covariance matrix of a subset of
 *  the dataset or the whole dataset.
 */

public class CovarianceCalculator {


  /*** member variables ***/
  private float[][] _data;
  private double[] _meanVector;      //stores the latest calculated mean values
  private SymmetricMatrix _covarianceMatrix;
  private int _dimension;




  //////////////////////////////////////////
  // CONSTRUCTOR                          //
  //////////////////////////////////////////
  /** Creates a CovarianceCalculator object.
   *
   */
  public CovarianceCalculator(float[][] data) {
    _data = data;
    calculateMembers();


  }

  /** Returns the arithmetic mean vector that was used for the calculation of the previously returned covariance matrix.
   *
   *  @return double array holding the arithmetic mean for each dimension.
   */
  public double[] getMeanVector() {
    return _meanVector;
  }

  /** Returns the classic covariance matrix for the given data.
  *  @return <code>dimNumber</code> x <code>dimNumber</code> SymmetricMatrix representing the classic covariance matrix.
  */
  public SymmetricMatrix getClassicCovarianceMatrix() {
    return _covarianceMatrix;
  }




  ////////////////////////////
  // PRIVATE MEMBERS        //
  ////////////////////////////
  private void calculateMembers() {
    initMembers();
    if (_data != null) {
      calculateMeanVector();
      calculateCovarianceMatrix();
    }
  }

  private void initMembers() {
    _meanVector = null;
    _covarianceMatrix = null;
    _dimension = 0;
    if (_data != null) _dimension = _data.length;
  }

  private void calculateMeanVector() {
    _meanVector = new double[_dimension];
    for (int dim = 0; dim < _dimension; dim++) {
      _meanVector[dim] = calculateMean(_data[dim]);
    }
  }


  // calculates the arithmetic mean of the given float array
  private double calculateMean(float[] data) {
    //1. make a check
    if (data == null) return 0;

    //2. create the mean
    int length = data.length;

    double sum = 0;
    for (int i = 0; i < length; i++) {
      sum += data[i];
    }
    return sum/length;

  }

  /**
  *  In the diagonale the variances of the dimensions are stored sum((x-mean(x))*(x-mean(x)))/(N-1)
  *  In the element x,y of the matrix the covariance between the dimensions x and y is stored sum((x-mean(x))*(y-mean(y)))/(N-1)
  **/
  private void calculateCovarianceMatrix() {
    double[][] covarianceMatrix = getZeroSquareMatrix(_dimension);




    //4. fill the covariance matrix
    int entryNum = _data[0].length;
    for (int i = 0; i < entryNum; i++) {

      //create the mean normalized entry => (x - mean(x))
      double[] meanNormalizedItem = getMeanNormalizedItem(i);

      //add to each entry in covariance matrix => sum((x-mean(x))*(y-mean(y)))
      for (int row = 0; row < _dimension; row++) {
        double meanNormalizedItemRow = meanNormalizedItem[row];
        for (int col = 0; col < _dimension; col++) {
          covarianceMatrix[row][col] += meanNormalizedItemRow * meanNormalizedItem[col];
        }
      }

    }


    //5. divide by N-1
    double divisor = entryNum-1;
    for (int row = 0; row < _dimension; row++) {
      for (int col = 0; col < _dimension; col++) {
        covarianceMatrix[row][col] /= divisor;
      }
    }

    _covarianceMatrix = getSymmetricMatrix(covarianceMatrix);


  }


  private double[][] getZeroSquareMatrix(int dimension) {
    double[][] zeroMatrix = new double[dimension][dimension];
    for (int i = 0; i < dimension; i++) {
      for (int j = 0; j < dimension; j++) {
        zeroMatrix[i][j] = 0;
      }
    }
    return zeroMatrix;
  }

  //create the mean normalized entry => (x - mean(x))
  private double[] getMeanNormalizedItem(int entryId) {
    double[] meanNormalizedItem = new double[_dimension];
    for (int dim = 0; dim < _dimension; dim++) {
      meanNormalizedItem[dim] = _data[dim][entryId] - _meanVector[dim];
    }
    return meanNormalizedItem;
  }


  private SymmetricMatrix getSymmetricMatrix(double[][] values) {
    try {
      return new SymmetricMatrix(values);
    }
    catch (Exception e) {
      e.printStackTrace();
      return null;
    }
  }









}
