package tools;

/*
 *  Copyright (C) 2018-2020 Juergen Platzer
 *
 *  This file is part of JInfoViz.
 *
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */


import java.util.Vector;


public class DistanceCalculator {

  private float[][] _data;
  private SymmetricMatrix _covarianceMatrix;
  private double[] _meanVector;

  private double[] _distances;
  private Vector<IndexAndValue> _indexAndDistances;

  ////////////////////////////////////
  // CONSTRUCTOR                    //
  ////////////////////////////////////
  public DistanceCalculator(float[][] data, SymmetricMatrix covarianceMatrix, double[] meanVector) {
    _data = data;
    _covarianceMatrix = covarianceMatrix;
    _meanVector = meanVector;

    _distances = null;
    _indexAndDistances = null;
  }


  public Vector<IndexAndValue> getMahalanobisDistances() throws Exception {
    if (_indexAndDistances == null) {
      calculateDistances();
    }

    return _indexAndDistances;
  }

  public double[] getMahalanobisDistancesAsArray() throws Exception {
    if (_distances == null) {
      calculateDistances();
    }

    return _distances;

  }





  //////////////////////////////////
  // PRIVATE METHODS              //
  //////////////////////////////////
  private void calculateDistances() throws Exception {
    if (_covarianceMatrix.isNotInvertable()) {
      throw new Exception("This matrix is not invertable!");
    }
    SquareMatrix inverseMatrix = _covarianceMatrix.getInverse();
    double[][] inverseCovarianceMatrix = inverseMatrix.getCopyOfValues();


    //3. calculate mahalanobis distance
    int dimNum = _data.length;
    int entryNum = _data[0].length;
    _indexAndDistances = new Vector<IndexAndValue>();
    _distances = new double[entryNum];

    for (int i = 0; i < entryNum; i++) {

      //3.1.create meannormalized entry
      double[] dataItem = new double[dimNum];
      for (int dim = 0; dim < dimNum; dim++) {
        dataItem[dim] = _data[dim][i] - _meanVector[dim];
      }

      //3.2. calculate the distance
      double distance = 0;
      for (int iRow = 0; iRow < dimNum; iRow++) {
        for (int iCol = 0; iCol < dimNum; iCol++) {
          distance += dataItem[iRow]*inverseCovarianceMatrix[iRow][iCol]*dataItem[iCol];
        }
      }

      distance = Math.sqrt(distance);
      if (Double.isNaN(distance)) {
        for (int dim = 0; dim < dimNum; dim++) {
          System.out.println("DataItem: " + dataItem[dim]);
        }

        _covarianceMatrix.printMatrixValues();

        System.out.println("det: " + _covarianceMatrix.getDeterminant());

        inverseMatrix.printMatrixValues();

        /*
        for (int iRow = 0; iRow < dimNum; iRow++) {
          for (int iCol = 0; iCol < dimNum; iCol++) {
            System.out.println("Matrix: " + (dataItem[iRow]*inverseCovarianceMatrix[iRow][iCol]*dataItem[iCol]));
          }
        }
        */

        throw new Exception("Distance is NaN!!!");

      }

      _indexAndDistances.add(new IndexAndValue(i, distance));
      _distances[i] = distance;

      //System.out.println((entry+1) + ": " + distance);
    }


  }



}
