package tools;

/*
 *  Copyright (C) 2020 Juergen Platzer
 *
 *  This file is part of JInfoViz.
 *
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.Vector;

public class EigenvalueDecomposition {

  /**
    Power-Iteration algorithm: https://en.wikipedia.org/wiki/Power_iteration
    https://de.wikipedia.org/wiki/Potenzmethode

    Eigenvectors:
     * Eigenvectors are vectors that do not change their direction by the matrix multiplication.
     * They are just elongated by their eigen value -> A*x = lambda*x, x is Eigenvector and lambda is Eigenvalue

    Idea of the algorithm.
     * By iteratively multiplying an arbitrary vector with the Matrix it converges to the eigenvector with the largest eigen value


    Finding other vectors: http://mlwiki.org/index.php/Power_Iteration#Naive_Method

    A = Q*lambda*Qt, Q is the matrix of eigenvectors and Qt the transposed.
    A = sum(lambda*qi*qti) for i from 1 to n
    1. find q1 and lambda 1
    2. remove dominant direction from Matrix A2 = A - lambda1*q1*qt1
    Repeat Power iteration and find lambda2 and q2

    x = c(1,2)

    x%*%t(x)
     [,1] [,2]
[1,]    1    2
[2,]    2    4


  **/

  private SymmetricMatrix _matrix;
  private int _dimension;
  private Vector<EigenValueAndVector> _eigenValuesAndVectors;

  private int MAX_ITERATIONS = 20;


  public  EigenvalueDecomposition(SymmetricMatrix matrix) throws Exception {
    _matrix = matrix;
    _dimension = matrix.getDimension();
    _eigenValuesAndVectors = new Vector<EigenValueAndVector>();

    calculateAllEigenVectorsAndValues();



  }

  public Vector<EigenValueAndVector> getEigenValuesAndVectors() {
    return _eigenValuesAndVectors;
  }

  public void print() {
    for (EigenValueAndVector eigenValAndVec : _eigenValuesAndVectors) {
      eigenValAndVec.print();
    }
  }






  ///////////////////////////////
  // PRIVATE METHODS           //
  ///////////////////////////////
  private void calculateAllEigenVectorsAndValues() throws Exception {
    SquareMatrix matrix = _matrix;
    for (int i = 0; i < _dimension; i++) {
      calculateNextEigenVectorAndValue(matrix);
      matrix = computeMatrixWithoutEigenVectorAndValue(matrix, _eigenValuesAndVectors.get(i));
    }

  }


  private void calculateNextEigenVectorAndValue(SquareMatrix matrix) throws Exception {
    double[] eigenVector = calculateNextEigenVector(matrix);
    double eigenValue = calculateEigenValueForVector(matrix, eigenVector);

    int id = _eigenValuesAndVectors.size()+1;
    _eigenValuesAndVectors.add(new EigenValueAndVector(id, eigenValue, eigenVector));

  }


  //By iteratively multiplying an arbitrary vector with the Matrix it converges to the eigenvector with the largest eigen value
  private double[] calculateNextEigenVector(SquareMatrix matrix) throws Exception  {

    double[] eigenVector = VectorUtilities.getRandomVector(_dimension);

    for (int i = 0; i < MAX_ITERATIONS; i++) {
      eigenVector = matrix.getVectorMultiplicationResult(eigenVector);
      eigenVector = VectorUtilities.normalizeVector(eigenVector);
    }
    return eigenVector;

  }

  //multiply the eigenvector with the matrix. The new length of the new vector is the eigenvalue A*x = lambda*x
  private double calculateEigenValueForVector(SquareMatrix matrix, double[] eigenVector) throws Exception {
    double[] elongatedEigenVector = matrix.getVectorMultiplicationResult(eigenVector);
    return VectorUtilities.getNormOfVector(elongatedEigenVector);
  }


  //remove dominant direction from Matrix A2 = A - lambda1*q1*qt1
  private SquareMatrix computeMatrixWithoutEigenVectorAndValue(SquareMatrix matrix, EigenValueAndVector eigenValueAndVector) throws Exception {

    SquareMatrix eigenVectorMatrix = VectorUtilities.getVectorMultipliedWithItsTransposed(eigenValueAndVector.getVector());
    eigenVectorMatrix.multiplyWithScalar(eigenValueAndVector.getValue());
    return matrix.getMatrixSubtractionResult(eigenVectorMatrix);


  }













}
