package tools;

/*
 *  Copyright (C) 2020 Juergen Platzer
 *
 *  This file is part of JInfoViz.
 *
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */

public class MCDBestSolutions {

  private double[] _bestSolutionValues;
  private MCDSolution[] _bestSolutions;
  private int _bestSolutionNumber;
  private int _solutionsFilled;


  public MCDBestSolutions(int bestSolutionNumber) {
    _bestSolutionNumber = bestSolutionNumber;
    if (_bestSolutionNumber < 2) _bestSolutionNumber = 2;


    _bestSolutionValues = new double[_bestSolutionNumber];
    _bestSolutions = new MCDSolution[_bestSolutionNumber];
    for (int i = 0; i < _bestSolutionNumber; i++) {
      _bestSolutionValues[i] = Double.POSITIVE_INFINITY;
      _bestSolutions[i] = null;
    }
    _solutionsFilled = 0;
  }


  public void updateBestSolutions(MCDSolution solution) {
    if (solution == null) return;
    if (_solutionsFilled < _bestSolutionNumber) {
      addSolution(solution);
    }
    else if (isBestSolution(solution)) {
      replaceWorstSolution(solution);
    }
  }

  public MCDSolution getBestSolution() {
    return _bestSolutions[getBestSolutionId()];
  }

  public MCDSolution[] getAllSolutions() {
    return _bestSolutions;
  }

  public void validate() {
    int errorCount = 0;
    for (int i = 0; i < _bestSolutionNumber; i++) {
      if (_bestSolutions[i] != null && _bestSolutionValues[i] != _bestSolutions[i].getSolutionValue()) {
        System.out.println("Value dismatch error in solution " + i + " solutionValue = " + _bestSolutions[i].getSolutionValue() + " vs. storedValue = " + _bestSolutionValues[i]);
        errorCount++;
      }
    }

    if (errorCount == 0) {
      System.out.println("Best Solutions is valid!");
    }
    else {
      System.out.println(errorCount + " errors found!");
    }
  }

  public void printBestSolutionValues() {
    for (int i = 0; i < _bestSolutionNumber; i++) {
      System.out.println("Solution " + i + ": value = " + _bestSolutionValues[i]);
    }
  }


  //////////////////////////////////
  // PRIVATE METHODS              //
  //////////////////////////////////

  private void addSolution(MCDSolution solution) {
    _bestSolutionValues[_solutionsFilled] = solution.getSolutionValue();
    _bestSolutions[_solutionsFilled] = solution;

    _solutionsFilled++;
  }


  private boolean isBestSolution(MCDSolution solution) {
    double solutionValue = solution.getSolutionValue();
    for (int i = 0; i < _bestSolutionNumber; i++) {
      if (solutionValue < _bestSolutionValues[i]) {
        return true;
      }
    }

    return false;
  }

  private void replaceWorstSolution(MCDSolution betterSolution) {
    int worstSolutionId = getWorstSolutionId();
    _bestSolutionValues[worstSolutionId] = betterSolution.getSolutionValue();
    _bestSolutions[worstSolutionId] = betterSolution;
  }

  private int getWorstSolutionId() {
    double worstValue = _bestSolutionValues[0];
    int worstSolutionId = 0;
    for (int i = 1; i < _bestSolutionNumber; i++) {
      if (_bestSolutionValues[i] > worstValue) {
        worstValue = _bestSolutionValues[i];
        worstSolutionId = i;
      }
    }
    return worstSolutionId;
  }

  private int getBestSolutionId() {
    double bestValue = _bestSolutionValues[0];
    int bestSolutionId = 0;
    for (int i = 1; i < _bestSolutionNumber; i++) {
      if (_bestSolutionValues[i] < bestValue) {
        bestValue = _bestSolutionValues[i];
        bestSolutionId = i;
      }
    }
    return bestSolutionId;
  }

}
