package tools;

/*
 *  Copyright (C) 2020 Juergen Platzer
 *
 *  This file is part of JInfoViz.
 *
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */


import java.util.Vector;
import java.util.Collections;



/** Class implementing the classic and robust calculation of the covariance matrix of a subset of
 *  the dataset or the whole dataset.
 */

public class MCDCovarianceCalculator {

  /*** constants ***/
  private static final int MCD_SUBSET_NUM = 500;
  private static final int MCD_INITIAL_SUBSET_ITERATIONS = 2;
  private static final int MCD_BEST_SOLUTION_NUM = 10;

  /*** member variables ***/
  private double _robustness;
  private float[][] _data;

  private int _dimension;
  private int _totalEntryNum;
  private int _fittedEntryNum;
  private CovarianceCalculator _bestCovCalculator;

  /*** MCD members ***/
  private Vector<Integer> _randomEntryIds;
  private float[][] _fittedData;
  private MCDBestSolutions _bestSolutions;

  //private double[] _meanVector;      //stores the latest calculated mean values




  //////////////////////////////////////////
  // CONSTRUCTOR                          //
  //////////////////////////////////////////
  /** Creates a CovarianceCalculator object.
   *
   */
  public MCDCovarianceCalculator(double robustness, float[][] data) {
    _robustness = robustness;
    _data = data;

    calculateMembers();

  }

  /** Returns the arithmetic mean vector that was used for the calculation of the previously returned covariance matrix.
   *
   *  @return double array holding the arithmetic mean for each dimension.
   */
  public double[] getRobustMeanVector() {
    return _bestCovCalculator.getMeanVector();
  }

  public SymmetricMatrix getRobustCovarianceMatrix() {
    return _bestCovCalculator.getClassicCovarianceMatrix();
  }



  //////////////////////////////
  // PRIVATE METHODS          //
  //////////////////////////////
  private void calculateMembers() {
    initMembers();
    if (_data != null) {
      if (_fittedEntryNum >= _totalEntryNum) {
        calculateClassicCovarianceMatrix();
      }
      else {
        calculateRobustCovarianceMatrix();
      }
    }

  }


  private void initMembers() {
    if (_robustness < 0.5) _robustness = 0.5;
    if (_robustness > 1) _robustness = 1;

    _dimension = 0;
    _totalEntryNum = 0;
    _fittedEntryNum = 0;
    _bestCovCalculator = null;

    if (_data != null) {
      _dimension = _data.length;
      _totalEntryNum = _data[0].length;
      _fittedEntryNum = (int)Math.floor((_totalEntryNum + _dimension + 1) * _robustness);
      if (_fittedEntryNum > _totalEntryNum) _fittedEntryNum = _totalEntryNum;
      //System.out.println("fitted entry num: " + _fittedEntryNum);
    }
  }

  private void calculateClassicCovarianceMatrix() {
    _bestCovCalculator = new CovarianceCalculator(_data);
  }


  //approximates the MCD (Minimum Covariance Determinant) covariance matrix
  //cov.mcd
  //library(MASS)
  // cov.mcd
  //cov.mcd(iris[1:4])
  /*
$center
Sepal.Length  Sepal.Width Petal.Length  Petal.Width
    6.201053     2.858947     4.824211     1.660000

$cov
             Sepal.Length Sepal.Width Petal.Length Petal.Width
Sepal.Length    0.3641478  0.10227772    0.3622083  0.14642553
Sepal.Width     0.1022777  0.09648824    0.1260045  0.07695745
Petal.Length    0.3622083  0.12600448    0.5714289  0.27023404
Petal.Width     0.1464255  0.07695745    0.2702340  0.17965957

$msg
[1] "9 singular samples of size 5 out of 2500"

$crit
[1] -9.811189

$best
 [1]  51  52  53  54  55  56  57  58  59  60  62  64  65  66  67  68  70  71  72
[20]  73  74  75  76  77  78  79  80  81  82  83  84  85  86  87  89  90  91  92
[39]  93  94  95  96  97  98 100 101 102 103 104 105 110 111 112 113 114 116 117
[58] 121 122 124 125 127 128 129 133 134 137 138 139 140 141 143 144 145 147 148
[77] 149 150

$n.obs
[1] 150
*/


  //choose 500 arbitrary subsets of the data of size fittedNum
  //make with each of those 500 subsets 2 iteration steps
  //keep the 10 bestsolutions
  //make for each of those 10 solutions iteration steps until the iteration converges
  //the best solution is the appoximation of the MCD covariance matrix

  /** Returns the robust covariance matrix for the given data, calculated according
   *  "A Fast Algorithm for the Minimum Covariance Determinant Estimator" by Peter J. Rousseeuw and Katrien Van Driessen.
   *
   *  @param robustness float value between 0.5 and 1 indicating on which percentage of data items should be used for
   *  the robust estimate of the covariance matrix.
   *  @param data a <code>dimNumber</code> x <code>entryNumber</code> float array holding holding the values
   *  of the data items in the relevant dimensions.
   *
   *  @return <code>dimNumber</code> x <code>dimNumber</code> float array representing the classic covariance matrix.
   */
  private void calculateRobustCovarianceMatrix() {
    initMCDMembers();
    findBestSubsets();
    improveBestSolutions();
    _bestCovCalculator = _bestSolutions.getBestSolution().getCovarianceCalculator();
  }


  private void initMCDMembers() {
    _randomEntryIds = new Vector<Integer>();
    for (int i = 0; i < _totalEntryNum; i++) {
      _randomEntryIds.add(new Integer(i));
    }

    _fittedData = new float[_dimension][_fittedEntryNum];
    _bestSolutions = new MCDBestSolutions(MCD_BEST_SOLUTION_NUM);

  }

  private void findBestSubsets() {
    for (int i = 0; i < MCD_SUBSET_NUM; i++) {

      chooseNextDataSubset();

      MCDSolution currentSolution = null;
      try {
        currentSolution = getInitialSolution();
      }
      catch (Exception e) {
        //current solution is invertable -> skip it
        continue;
      }

      _bestSolutions.updateBestSolutions(currentSolution);
    }

  }

  private void chooseNextDataSubset() {
    Collections.shuffle(_randomEntryIds);

    //7.2. fill the fittedData with the first fittedNum IDs of the randomNumber vector
    for (int i = 0; i < _fittedEntryNum; i++) {
      int entryId = (_randomEntryIds.get(i)).intValue();

      for (int dim = 0; dim < _dimension; dim++) {
        _fittedData[dim][i] = _data[dim][i];
      }
    }
  }

  private MCDSolution getInitialSolution() throws Exception {
    CovarianceCalculator covCalculator = new CovarianceCalculator(_fittedData);
    SymmetricMatrix covarianceMatrix = covCalculator.getClassicCovarianceMatrix();
    if (covarianceMatrix.isNotInvertable()) {
      throw new Exception("This initial solution has an invertable Covariance matrix");
    }

    return improveSolution(new MCDSolution(covCalculator), MCD_INITIAL_SUBSET_ITERATIONS);
  }

  private void improveBestSolutions() {
    MCDSolution[] solutions = _bestSolutions.getAllSolutions();
    int solutionNum = solutions.length;
    for (int i = 0; i < solutionNum; i++) {
      try {
        MCDSolution newSolution = improveSolution(solutions[i], -1);
        _bestSolutions.updateBestSolutions(newSolution);
      }
      catch (Exception e) {}

    }

  }






  private MCDSolution improveSolution(MCDSolution solution, int maxIterationSteps) throws Exception {

    //2. make the basic settings for the improvement loop
    boolean continueImprovement = true;

    CovarianceCalculator bestCovCalculator = solution.getCovarianceCalculator();
    SymmetricMatrix bestMatrix = bestCovCalculator.getClassicCovarianceMatrix();
    if (bestMatrix.isNotInvertable()) {
      throw new Exception("The initial covariance matrix is not invertable!");
    }
    double minDeterminant = bestMatrix.getDeterminant();

    float[][] fittedData = null;
    int counter = 0;


    //3. iterative improve the solution
    while (continueImprovement) {

      //3.1. get the data points with the smallest mahalanobis distance for the current solution
      try {
        fittedData = getSmallestDistancePoints(bestCovCalculator);
      }
      catch (Exception e) {
        return new MCDSolution(bestCovCalculator);
      }


      //3.2. get classic covariance matrix for those points
      CovarianceCalculator newCovCalculator = new CovarianceCalculator(fittedData);
      SymmetricMatrix newCovarianceMatrix = newCovCalculator.getClassicCovarianceMatrix();
      if (newCovarianceMatrix.isNotInvertable()) {
        return new MCDSolution(bestCovCalculator);
      }
      double newDeterminant = newCovarianceMatrix.getDeterminant();

      //3.4. if this was no improvement -> stop the run, else set the improved solution and go on
      if (newDeterminant >= minDeterminant) {
        continueImprovement = false;
      }
      else {
        bestCovCalculator = newCovCalculator;
        minDeterminant = newDeterminant;
      }

      //3.5. check if the maximum iteration num has been reached.
      //    (maxIterationSteps == -1 -> there is no maximum iteration depth)
      counter++;

      if ((maxIterationSteps > 0) && (counter >= maxIterationSteps)) {
        continueImprovement = false;
      }

      //System.out.println("Log(det) = " + Math.log(fMinDet));

    }


    return new MCDSolution(bestCovCalculator);


  }






  //calculates the mahalanobis distance for each datapoint
  //returns the index of the n data points that have the smallest mahalanobis distance
  private float[][] getSmallestDistancePoints(CovarianceCalculator covCalculator) throws Exception {




    //2. calculate mahalanobis distance
    DistanceCalculator distCalc = new DistanceCalculator(_data, covCalculator.getClassicCovarianceMatrix(), covCalculator.getMeanVector());
    Vector<IndexAndValue> distances = distCalc.getMahalanobisDistances();

    //3. sort the distances
    Collections.sort(distances);


    //4. take the first n entries and return them
    float[][] smallestDistancePoints = new float[_dimension][_fittedEntryNum];
    for (int i = 0; i < _fittedEntryNum; i++) {
      int entryId = distances.get(i).getIndex();
      for (int dim = 0; dim < _dimension; dim++) {
        smallestDistancePoints[dim][i] = _data[dim][entryId];
      }

    }

    return smallestDistancePoints;

  }



}
