package tools;

/*
 *  Copyright (C) 2020 Juergen Platzer
 *
 *  This file is part of JInfoViz.
 *
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */

public class MatrixInversion {



/**
 Idea of the algorithm: Modify the matrix with operations so that it becomes the identity and apply the same operations to the
 identity matrix. The transformed identity matrix is the inverse.


 Example:

 1   2   |   1   0
 2   1   |   0   1

 1. New row 2 = Row 2 - (2/1)*Row1 (ie. make all values zero under the diagonal)

 1   2   |   1   0
 0  -3   |  -2   1

 2. New row 1 = Row1 - (-2/3)*Row2  (ie. make all values zero above of the diagonal)

 1   0   |   1-4/3   2/3
 0  -3   |  -2       1

 3. Normalize each row
 1   0   |   -1/3   2/3
 0   1   |   2/3    -1/3


The determinant is the product of the diagonal of the matrix before the rows have been normalized

 *
 **/

  private SymmetricMatrix _matrix;
  private double[][] _matrixValues;
  private double[][] _inverse;
  private int _dimension;
  private double _determinant;
  private boolean _minDivisorReached;

  private static final double MIN_DETERMINANT_VALUE = 0.0000000001;
  private static final double MIN_DIVISOR_VALUE = 0.0001;

  public MatrixInversion(SymmetricMatrix matrix) throws Exception {
    _matrix = matrix;
    _matrixValues = matrix.getCopyOfValues();
    _dimension = matrix.getDimension();

    calculateInverseAndDeterminant();
  }




  public SquareMatrix getInverse() throws Exception {
    return new SquareMatrix(_inverse);
  }

  public double getDeterminant() {
    return _determinant;
  }

  //when determinant nearly 0
  public boolean isNotInvertable() {
    return Double.isNaN(_determinant) || (Math.abs(_determinant) < MIN_DETERMINANT_VALUE) || _minDivisorReached;
  }


  ////////////////////////////////////
  // PRIVATE METHODS                //
  ////////////////////////////////////
  private void calculateInverseAndDeterminant() {
    initInverseMatrix();
    createZerosUnderTheDiagonal();
    createZerosAboveTheDiagonal();
    computeDeterminant();
    normalizeRows();
  }

  private void initInverseMatrix() {
    _inverse = getIdentityMatrix(_dimension);
    _minDivisorReached = false;
  }

  private void createZerosUnderTheDiagonal() {
    for (int col = 0; col < _dimension; col++) {
      createZerosUnderTheDiagonalOfColumn(col);
    }
  }

  private void createZerosUnderTheDiagonalOfColumn(int column) {
    double divisor = _matrixValues[column][column];
    checkDivisor(divisor);

    for (int row = column+1; row < _dimension; row++) {
      double factor = _matrixValues[row][column]/divisor;
      //System.out.println("factor: " + factor);
      subtractRow(_matrixValues, row, column, factor);
      subtractRow(_inverse, row, column, factor);
    }
  }


  private void createZerosAboveTheDiagonal() {
    for (int col = _dimension-1; col >= 0; col--) {
      createZerosAboveTheDiagonalOfColumn(col);
    }
  }

  private void createZerosAboveTheDiagonalOfColumn(int column) {
    double divisor = _matrixValues[column][column];
    checkDivisor(divisor);

    for (int row = column-1; row >= 0; row--) {
      double factor = _matrixValues[row][column]/divisor;
      subtractRow(_matrixValues, row, column, factor);
      subtractRow(_inverse, row, column, factor);
    }
  }

  private void normalizeRows() {
    for (int i = 0; i < _dimension; i++) {
      double divisor = _matrixValues[i][i];
      normailzeRow(_matrixValues, i, divisor);
      normailzeRow(_inverse, i, divisor);
    }
  }

  private void checkDivisor(double divisor) {
    if (Math.abs(divisor) < MIN_DIVISOR_VALUE) {
      _minDivisorReached = true;
    }
  }


  private void subtractRow(double[][] matrix, int row1, int row2, double factor) {
    for (int col = 0; col < _dimension; col++) {
      matrix[row1][col] -= matrix[row2][col]*factor;
    }
  }

  private void normailzeRow(double[][] matrix, int row, double divisor) {
    for (int col = 0; col < _dimension; col++) {
      matrix[row][col] /= divisor;
    }
  }

  private void computeDeterminant() {
    _determinant = 1;
    for (int i = 0; i < _dimension; i++) {
      _determinant *= _matrixValues[i][i];
    }
  }

  //TODO: generic matrix method?
  private static double[][] getIdentityMatrix(int dimension) {
    if (dimension < 2) dimension = 2;

    double[][] identity = new double[dimension][dimension];
    for (int i = 0; i < dimension; i++) {
      for (int j = 0; j < dimension; j++) {
        double value = 0;
        if (i == j) value = 1;
        identity[i][j] = value;
      }
    }
    return identity;
  }





}
