package tools;

/*
 *  Copyright (C) 2018-2020 Juergen Platzer
 *
 *  This file is part of JInfoViz.
 *
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.*;

public class PrincipalComponentAnalysis {


  private int       _dimNum;
  private int       _entryNum;
  private float[][] _data; //TODO: switch this to double

  private Vector<EigenValueAndVector>    _principalComponents;
  private double[] _componentVariances;
  private double     _sumOfVariances;

  private double[]   _usedMeans;
  private Vector<double[]> _componentCoordinates;

  private double     _robustness;                    //robustness parameter

  /////////////////////////////////
  // CONSTRUCTOR                 //
  /////////////////////////////////
  //constructor for classic pca
  public PrincipalComponentAnalysis(float[][] pData, double robustness) {
    initMembers();
    setRobustness(robustness);
    setData(pData);
  }


  /////////////////////////////////
  // SET METHODS                 //
  /////////////////////////////////
  public void setRobustness(double robustness) {
    if (robustness >= 1) robustness = 1;
    if (robustness <= 0.5) robustness = 0.5;

    boolean recalculatePCA = (_robustness != robustness);
    _robustness = robustness;
    if (recalculatePCA) calculatePrincipalComponents();

  }

  public void setData(float[][] data) {
    _data = data;
    _dimNum = data.length;
    _entryNum = data[0].length;
    calculatePrincipalComponents();
  }



  /////////////////////////////////
  // GET METHODS                 //
  /////////////////////////////////

  //returns the explained variance by the components that are specified by an index array
  public double getExplainedVarianceByComponents(int[] indices) {
    //1. make some checks
    if (indices == null) return 0;
    int indicesNum = indices.length;
    if (indicesNum == 0) return 0;

    //2. calculate the value and return it
    double explainedVariance = 0;

    for (int i = 0; i < indicesNum; i++) {
      explainedVariance += getExplainedVarianceByComponent(indices[i]);
    }

    return explainedVariance;

  }


  //returns the explained variance by the component that is specified by index (1 = 1. principal component)
  public double getExplainedVarianceByComponent(int componentIndex) {

    //EigenValueAndVector principalComponent = getPrincipalComponent(componentIndex);
    //if (principalComponent == null) return 0;
    int entryId = mapComponentIndexToInternalId(componentIndex);
    //System.out.println(_componentVariances[entryId]);
    return (_componentVariances[entryId]/_sumOfVariances);

  }




  //returns for data used for creation of the pca the coordinates in the specified principal component
  public double[] getComponentCoordinates(int componentIndex) {

    int entryId = mapComponentIndexToInternalId(componentIndex);
    return _componentCoordinates.get(entryId);

    /*
    //get the needed principal component
    EigenValueAndVector principalComponent =getPrincipalComponent(componentIndex);
    if (principalComponent == null) return null;
    double[] principalComponentVector = principalComponent.getVector();

    //calculate the entry coordinates on the component
    double[] componentCoordinates = new double[_entryNum];

    for (int i = 0; i < _entryNum; i++) {
      double entryCoordinatesOnPrincipalComponent = 0;
      for (int dim = 0; dim < _dimNum; dim++) {
        entryCoordinatesOnPrincipalComponent += (_data[dim][i] - _usedMeans[dim]) * principalComponentVector[dim];
      }
      componentCoordinates[i] = entryCoordinatesOnPrincipalComponent;
    }

    return componentCoordinates;
    */


  }


  public EigenValueAndVector getPrincipalComponent(int componentIndex) {

    //make some checks
    int entryId = mapComponentIndexToInternalId(componentIndex);
    if (entryId < 0) return null;
    return _principalComponents.get(entryId);
  }



  public double getRobustness() {
    return _robustness;
  }

  public int getPrincipalComponentNumber() {
    return _principalComponents.size();
  }



  ////////////////////////////////////
  // PRIVATE METHODS                //
  ////////////////////////////////////
  private void initMembers() {
    _dimNum = 0;
    _entryNum = 0;
    _data = null;
    _principalComponents = null;
    _sumOfVariances = 0;
    _usedMeans = null;
    _robustness = 1;
  }


  private void calculatePrincipalComponents() {
    if (_data == null) return;

    SymmetricMatrix covarianceMatrix = getCovarianceMatrix();
    setPrincipalComponents(covarianceMatrix);
    calculateMeans();
    calculateComponentCoordinates();
    calculateVariances();

  }

  private SymmetricMatrix getCovarianceMatrix() {
    if (_robustness == 1) {
      CovarianceCalculator covCalculator = new CovarianceCalculator(_data);
      return covCalculator.getClassicCovarianceMatrix();
    }

    try {
      MCDCovarianceCalculator robustCovCalculator = new MCDCovarianceCalculator(_robustness, _data);
      return robustCovCalculator.getRobustCovarianceMatrix();
    }
    catch (Exception e) {
      System.out.println("Robust covariance matrix cannot be calculated!");
      e.printStackTrace();
      CovarianceCalculator covCalculator = new CovarianceCalculator(_data);
      return covCalculator.getClassicCovarianceMatrix();
    }
  }

  private void setPrincipalComponents(SymmetricMatrix covarianceMatrix) {
    try {
      _principalComponents = covarianceMatrix.getEigenValuesAndVectors();
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }




  private void calculateMeans() {
    _usedMeans = new double[_dimNum];
    for (int dim = 0; dim < _dimNum; dim++) {
      _usedMeans[dim] = calculateMean(_data[dim]);
    }
  }

  //TODO: this should be part of the data dimensions (also in the covariance calculator this is implemented!)
  // calculates the arithmetic mean of the given float array
  private double calculateMean(float[] data) {
    //1. make some checks
    if (data == null) return 0;

    //2. do the calculation
    int entryNum = data.length;

    double sum = 0;
    for (int i = 0; i < entryNum; i++) {
      sum += data[i];
    }
    return sum/entryNum;

  }

  private void calculateComponentCoordinates() {
    _componentCoordinates = new Vector<double[]>();
    for (int i = 0; i < _dimNum; i++) {
      double[] principalComponentVector = _principalComponents.get(i).getVector();
      _componentCoordinates.add(calculateCoordinatesForComponent(principalComponentVector));
    }
  }

  private double[] calculateCoordinatesForComponent(double[] principalComponentVector) {

    //calculate the entry coordinates on the component
    double[] componentCoordinates = new double[_entryNum];

    for (int i = 0; i < _entryNum; i++) {
      double entryCoordinatesOnPrincipalComponent = 0;
      for (int dim = 0; dim < _dimNum; dim++) {
        entryCoordinatesOnPrincipalComponent += (_data[dim][i] - _usedMeans[dim]) * principalComponentVector[dim];
      }
      componentCoordinates[i] = entryCoordinatesOnPrincipalComponent;
    }
    return componentCoordinates;
  }

  private void calculateVariances() {
    _sumOfVariances = 0;
    _componentVariances = new double[_dimNum];
    for (int i = 0; i < _dimNum; i++) {
      double variance = calculateComponentVariance(i);
      _componentVariances[i] = variance;
      _sumOfVariances += variance;
    }

  }

  private double calculateComponentVariance(int componentId) {
    double variance = 0;
    double[] coordinates = _componentCoordinates.get(componentId);
    for (int i = 0; i < _entryNum; i++) {
      variance += coordinates[i]*coordinates[i];
    }
    return variance/(_entryNum-1);
  }

  private int mapComponentIndexToInternalId(int componentIndex) {
    int entryId = componentIndex-1;

    if (entryId < 0) return -1;
    if (entryId >= _principalComponents.size()) return -1;
    return entryId;
  }

}
