package tools;

/*
 *  Copyright (C) 2020 Juergen Platzer
 *
 *  This file is part of JInfoViz.
 *
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */


import java.text.DecimalFormat;

public class SquareMatrix {


  private double[][] _values;
  private final static DecimalFormat NUMBER_FORMAT = new DecimalFormat("#0.000");

  public SquareMatrix(int dimension) {
    if (dimension < 2) dimension = 2;
    _values = new double[dimension][dimension];
    fillAsUnitMatrix();
  }

  public SquareMatrix(double[][] values) throws Exception {
    int dimension = values.length;
    if (dimension <= 1) throw new Exception("The matrix must have at least dimension 2!");
    if (values[0].length != dimension) throw new Exception("The matrix is not a square matrix!");
    _values = getCopyOfValues(values);
  }

  public SquareMatrix(float[][] values) throws Exception {
    int dimension = values.length;
    if (dimension <= 1) throw new Exception("The matrix must have at least dimension 2!");
    if (values[0].length != dimension) throw new Exception("The matrix is not a square matrix!");
    _values = getCopyOfValues(values);
  }


  //////////////////////////////
  // SET METHODS              //
  //////////////////////////////
  public void setValue(int i, int j, double value) {
    _values[i][j] = value;
  }

  //////////////////////////////
  // GET METHODS              //
  //////////////////////////////
  public int getDimension() {
    return _values.length;
  }

  public double[][] getCopyOfValues() {
    return getCopyOfValues(_values);
  }

  public float[][] getCopyOfValuesAsFloat() {
    return getCopyOfValuesAsFloat(_values);
  }

  public void printMatrixValues() {
    int dim = getDimension();
    System.out.println("\n" + "Matrix (" + dim + "x" + dim + ") = ");

    for (int i = 0; i < dim; i++) {
      String lineValues = "";
      for (int j = 0; j < dim; j++) {
        lineValues += NUMBER_FORMAT.format(_values[i][j]);
        lineValues += "   ";
      }
      System.out.println(lineValues);
    }


  }

  //////////////////////////////
  // OPERATION METHODS        //
  //////////////////////////////
  public double[] getVectorMultiplicationResult(double[] vector) throws Exception {
    if (vector == null) throw new NullPointerException("No vector has been specified!");
    int dim = getDimension();
    if (vector.length != dim) throw new Exception("The vector does not have the requested dimension " + dim + "!");

    double[] resultVector = new double[dim];

    for (int i = 0; i < dim; i++) {
      double productSum = 0;
      for (int j = 0; j < dim; j++) {
        productSum += _values[i][j]*vector[j];
      }
      resultVector[i] = productSum;
    }

    return resultVector;

  }

  public SquareMatrix getMatrixSubtractionResult(SquareMatrix matrixToSubtract) throws Exception {
    int dim = getDimension();
    if (dim != matrixToSubtract.getDimension()) throw new Exception("Matrix has a different dimension!");

    double[][] values1 = getCopyOfValues();
    double[][] values2 = matrixToSubtract.getCopyOfValues();
    for (int i = 0; i < dim; i++) {
      for (int j = 0; j < dim; j++) {
        values1[i][j] -= values2[i][j];
      }
    }

    return new SquareMatrix(values1);


  }

  public void multiplyWithScalar(double scalar) {
    int dim = getDimension();
    for (int i = 0; i < dim; i++) {
      for (int j = 0; j < dim; j++) {
        _values[i][j] *= scalar;
      }
    }
  }


  ///////////////////////////////////
  // PRIVATE METHODS               //
  ///////////////////////////////////
  private void fillAsUnitMatrix() {
    int dim = _values.length;
    for (int i = 0; i < dim; i++) {
      for (int j = 0; j < dim; j++) {
        double value = 0;
        if (i == j) value = 1;
        _values[i][j] = value;

      }
    }
  }



  private double[][] getCopyOfValues(double[][] origValues) {
    int dim = origValues.length;
    double[][] values = new double[dim][dim];

    for (int i = 0; i < dim; i++) {
      for (int j = 0; j < dim; j++) {
        values[i][j] = origValues[i][j];
      }
    }

    return values;
  }

  private double[][] getCopyOfValues(float[][] origValues) {
    int dim = origValues.length;
    double[][] values = new double[dim][dim];

    for (int i = 0; i < dim; i++) {
      for (int j = 0; j < dim; j++) {
        values[i][j] = (double)origValues[i][j];
      }
    }

    return values;
  }

  private float[][] getCopyOfValuesAsFloat(double[][] origValues) {
    int dim = origValues.length;
    float[][] values = new float[dim][dim];

    for (int i = 0; i < dim; i++) {
      for (int j = 0; j < dim; j++) {
        values[i][j] = (float)origValues[i][j];
      }
    }

    return values;
  }


}
